@@ -7,6 +7,7 @@
 #
 ##############################################################################
 
+import copy
 import logging
 from collections import namedtuple
 from httplib import BAD_REQUEST, UNAUTHORIZED, FORBIDDEN, OK
@@ -51,12 +52,13 @@ from .util import (
 from .shell import (
     _find_shell_id,
     _build_command_line_elem,
+    _build_ps_command_line_elem,
     _find_command_id,
     _MAX_REQUESTS_PER_COMMAND,
     _find_stream,
     _find_exit_code,
     CommandResponse,
-    _stripped_lines
+    _stripped_lines,
 )
 from .enumerate import (
     DEFAULT_RESOURCE_URI,
@@ -67,48 +69,6 @@ from .SessionManager import SESSION_MANAGER, Session
 kerberos = None
 LOG = logging.getLogger('winrm')
 
-EnumInfo = namedtuple('EnumInfo', ['wql', 'resource_uri'])
-
-
-def _build_ps_command_line_elem(ps_command, ps_script):
-    """Build PowerShell command line elements without splitting
-    the actual ps script into arguments. using _build_command_line_elem
-    with a ps script splits the script into separate arguments.  Remote
-    Windows shell inserts spaces when reconstituting the script.
-
-    ps_command - powershell command with arguments as string
-        e.g. 'powershell -NoLogo -NonInteractive -NoProfile -Command'
-    ps_script - script to be run in powershell as single line string
-        e.g. "& {get-counter -counter \"\memory\pages output/sec\" }"
-    """
-    command_line_parts = shlex.split(ps_command, posix=False)
-    # ensure '-command' is last
-    if command_line_parts[-1:][0].lower() != '-command':
-        index = 0
-        for option in command_line_parts:
-            if option.lower() == '-command':
-                command_line_parts.pop(index)
-                break
-            index += 1
-        command_line_parts.append(option)
-    prefix = "rsp"
-    ET.register_namespace(prefix, c.XML_NS_MSRSP)
-    command_line_elem = ET.Element('{%s}CommandLine' % c.XML_NS_MSRSP)
-    command_elem = ET.Element('{%s}Command' % c.XML_NS_MSRSP)
-    command_elem.text = command_line_parts[0]
-    command_line_elem.append(command_elem)
-    for arguments_text in command_line_parts[1:]:
-        arguments_elem = ET.Element('{%s}Arguments' % c.XML_NS_MSRSP)
-        arguments_elem.text = arguments_text
-        command_line_elem.append(arguments_elem)
-    arguments_elem = ET.Element('{%s}Arguments' % c.XML_NS_MSRSP)
-    arguments_elem.text = ps_script
-    command_line_elem.append(arguments_elem)
-    tree = ET.ElementTree(command_line_elem)
-    str_io = StringIO()
-    tree.write(str_io, encoding='utf-8')
-    return str_io.getvalue()
-
 
 class WinRMSession(Session):
     '''
@@ -250,7 +210,7 @@ class WinRMSession(Session):
                 import xml.dom.minidom
                 xml = xml.dom.minidom.parseString(xml_str)
                 LOG.debug(xml.toprettyxml())
-            except:
+            except Exception:
                 LOG.debug('Could not prettify response XML: "{0}"'.format(xml_str))
         returnValue(ET.fromstring(xml_str))
 
@@ -305,49 +265,62 @@ class WinRMClient(object):
         verify_conn_info(conn_info)
         self.key = None
         self._conn_info = conn_info
-        self.session_manager = SESSION_MANAGER
-        self._session = None
         self.ps_script = None
 
+    def _session(self):
+        return SESSION_MANAGER.get_connection(self.key)
+
     @inlineCallbacks
     def init_connection(self):
-        '''Initialize a connection through the session_manager'''
-        yield self.session_manager.init_connection(self, WinRMSession)
-        self._session = self.session_manager.get_connection(self.key)
+        """Initialize a connection through the session_manager"""
+        if self._session() is not None:
+            try:
+                self._refresh_dc.cancel()
+            except Exception:
+                pass
+        yield SESSION_MANAGER.init_connection(self, WinRMSession)
         returnValue(None)
 
     def is_kerberos(self):
         return self._conn_info.auth_type == 'kerberos'
 
     def decrypt_body(self, body):
-        return self._session.decrypt_body(body)
+        return self._session().decrypt_body(body)
+
+    @inlineCallbacks
+    def send_request(self, request, **kwargs):
+        if not self._session():
+            yield self.init_connection()
+
+        if not self._session():
+            raise Exception('Could not connect to device {}'.format(self.conn_info.hostname))
+        response = yield self._session().send_request(request, self, **kwargs)
+        returnValue(response)
 
     @inlineCallbacks
     def _create_shell(self):
-        elem = yield self._session.send_request('create', self)
+        elem = yield self.send_request('create')
         returnValue(_find_shell_id(elem))
 
     @inlineCallbacks
     def _delete_shell(self, shell_id):
-        yield self._session.send_request('delete', self, shell_id=shell_id)
+        yield self.send_request('delete', shell_id=shell_id)
         returnValue(None)
 
     @inlineCallbacks
     def _signal_terminate(self, shell_id, command_id):
-        yield self._session.send_request('signal',
-                                         self,
-                                         shell_id=shell_id,
-                                         command_id=command_id,
-                                         signal_code=c.SHELL_SIGNAL_TERMINATE)
+        yield self.send_request('signal',
+                                shell_id=shell_id,
+                                command_id=command_id,
+                                signal_code=c.SHELL_SIGNAL_TERMINATE)
         returnValue(None)
 
     @inlineCallbacks
     def _signal_ctrl_c(self, shell_id, command_id):
-        yield self._session.send_request('signal',
-                                         self,
-                                         shell_id=shell_id,
-                                         command_id=command_id,
-                                         signal_code=c.SHELL_SIGNAL_CTRL_C)
+        yield self.send_request('signal',
+                                shell_id=shell_id,
+                                command_id=command_id,
+                                signal_code=c.SHELL_SIGNAL_CTRL_C)
         returnValue(None)
 
     @inlineCallbacks
@@ -360,20 +333,20 @@ class WinRMClient(object):
         LOG.debug('WinRMClient._send_command: sending command request '
                   '(shell_id={0}, command_line_elem={1})'.format(
                       shell_id, command_line_elem))
-        command_elem = yield self._session.send_request(
-            'command', self, shell_id=shell_id, command_line_elem=command_line_elem,
+        command_elem = yield self.send_request(
+            'command', shell_id=shell_id, command_line_elem=command_line_elem,
             timeout=self._conn_info.timeout)
         returnValue(command_elem)
 
     @inlineCallbacks
     def _send_receive(self, shell_id, command_id):
-        receive_elem = yield self._session.send_request(
-            'receive', self, shell_id=shell_id, command_id=command_id)
+        receive_elem = yield self.send_request(
+            'receive', shell_id=shell_id, command_id=command_id)
         returnValue(receive_elem)
 
     @inlineCallbacks
     def close_connection(self):
-        yield self._session.close_connection(self)
+        yield self._session().close_connection(self)
         returnValue(None)
 
 
@@ -397,8 +370,8 @@ class SingleCommandClient(WinRMClient):
         self.ps_script = ps_script
         yield self.init_connection()
         try:
-            cmd_response = yield self._session.sem.run(self.run_single_command,
-                                                       command_line)
+            cmd_response = yield self._session().sem.run(self.run_single_command,
+                                                         command_line)
         except Exception:
             yield self.close_connection()
             raise
@@ -449,7 +422,14 @@ class SingleCommandClient(WinRMClient):
 
 
 class LongCommandClient(WinRMClient):
-    """Client to run a single long running command to a winrm device"""
+    """UNDER CONSTRUCTION!!! DO NOT USE!!
+
+    Having issues with this client.  May Need to be reworked a bit
+
+    TODO: If raising exception, we need to delete the remote shell
+    before closing connection
+
+    Client to run a single long running command to a winrm device"""
     def __init__(self, conn_info):
         super(LongCommandClient, self).__init__(conn_info)
         self._shell_id = None
@@ -470,11 +450,13 @@ class LongCommandClient(WinRMClient):
         self.key = (self._conn_info.ipaddress, command_line + str(ps_script))
         self.ps_script = ps_script
         yield self.init_connection()
-        self._shell_id = yield self._create_shell()
+        if self._shell_id is None:
+            self._shell_id = yield self._create_shell()
         try:
             command_elem = yield self._send_command(self._shell_id,
                                                     command_line)
         except TimeoutError:
+            yield self._sender.send_request('delete', shell_id=self._shell_id)
             yield self.close_connection()
             raise
         self._command_id = _find_command_id(command_elem)
@@ -524,7 +506,8 @@ class EnumerateClient(WinRMClient):
     @inlineCallbacks
     def enumerate(self, wql, resource_uri=DEFAULT_RESOURCE_URI):
         """Runs a remote WQL query."""
-        yield self.init_connection()
+        if self._session() is None:
+            yield self.init_connection()
         request_template_name = 'enumerate'
         enumeration_context = None
         items = []
@@ -532,7 +515,7 @@ class EnumerateClient(WinRMClient):
             for i in xrange(_MAX_REQUESTS_PER_ENUMERATION):
                 LOG.info('{0} "{1}" {2}'.format(
                     self._hostname, wql, request_template_name))
-                response = yield self._session._send_request(
+                response = yield self._session()._send_request(
                     request_template_name,
                     self,
                     resource_uri=resource_uri,
@@ -564,12 +547,11 @@ class EnumerateClient(WinRMClient):
         """
         items = {}
         yield self.init_connection()
-        self._session = self.session_manager.get_connection(self.key)
         for enum_info in enum_infos:
             try:
-                items[enum_info] = yield self._session.sem.run(self.enumerate,
-                                                               enum_info.wql,
-                                                               enum_info.resource_uri)
+                items[enum_info] = yield self._session().sem.run(self.enumerate,
+                                                                 enum_info.wql,
+                                                                 enum_info.resource_uri)
             except (UnauthorizedError, ForbiddenError):
                 # Fail the collection for general errors.
                 raise
@@ -628,17 +610,23 @@ class AssociatorClient(EnumerateClient):
 
         see https://msdn.microsoft.com/en-us/library/aa384793(v=vs.85).aspx
         """
+        associations_copy = copy.deepcopy(associations)
         items = {}
         wql = 'Select {} from {}'.format(','.join(fields), seed_class)
         if where:
             wql += ' where {}'.format(where)
+        EnumInfo = namedtuple('EnumInfo', ['wql', 'resource_uri'])
         enum_info = EnumInfo(wql, resource_uri)
         results = yield self.do_collect([enum_info])
-        input_results = results[enum_info]
+
+        try:
+            input_results = results[enum_info]
+        except KeyError:
+            raise Exception('No results for seed class {}.'.format(seed_class))
 
         items[seed_class] = input_results
-        while associations:
-            association = associations.pop(0)
+        while associations_copy:
+            association = associations_copy.pop(0)
             associate_results = []
             prop_results = {}
             for item in input_results:
