@@ -19,6 +19,7 @@ from xml.etree.ElementTree import ParseError
 from twisted.internet import reactor, defer
 from twisted.internet.protocol import Protocol
 from twisted.web.client import Agent
+from twisted.internet.error import ConnectError
 from twisted.internet.ssl import ClientContextFactory
 from twisted.web.http_headers import Headers
 from twisted.internet.threads import deferToThread
@@ -149,7 +150,11 @@ class _ErrorReader(Protocol):
 
     def connectionLost(self, reason):
         if self.gssclient:
-            body = self.gssclient.decrypt_body(''.join(self._data))
+            try:
+                body = self.gssclient.decrypt_body(''.join(self._data))
+            except Exception as e:
+                body = 'There was a problem decrypting an error message: {}.'\
+                       ' Check WinRM logs on {}'.format(e.message, self.gssclient._conn_info.hostname)
         else:
             body = ''.join(self._data)
         message = _parse_error_message(body)
@@ -265,7 +270,8 @@ class AuthGSSClient(object):
                 kinit_result = yield kinit(self._username,
                                            self._password,
                                            self._dcip,
-                                           includedir=self._include_dir)
+                                           includedir=self._include_dir,
+                                           disable_rdns=self._conn_info.disable_rdns)
                 if kinit_result:
                     # this error is ok.  it just means more
                     # than one process is calling kinit
@@ -432,11 +438,12 @@ class ConnectionInfo(namedtuple(
         'envelope_size',
         'code_page',
         'locale',
-        'include_dir'])):
+        'include_dir',
+        'disable_rdns'])):
     def __new__(cls, hostname, auth_type, username, password, scheme, port,
                 connectiontype, keytab, dcip, timeout=60, trusted_realm='',
                 trusted_kdc='', ipaddress='', service='', envelope_size=512000,
-                code_page=65001, locale='en-US', include_dir=None):
+                code_page=65001, locale='en-US', include_dir=None, disable_rdns=False):
         if not ipaddress:
             ipaddress = hostname
         if not service:
@@ -448,7 +455,7 @@ class ConnectionInfo(namedtuple(
                                                   trusted_realm, trusted_kdc,
                                                   ipaddress, service,
                                                   envelope_size, code_page, locale,
-                                                  include_dir)
+                                                  include_dir, disable_rdns)
 
 
 def verify_include_dir(conn_info):
@@ -614,30 +621,43 @@ class RequestSender(object):
             body_producer = _StringProducer(encrypted_request)
         else:
             body_producer = _StringProducer(request)
+
+        @defer.inlineCallbacks
+        def reset_agent_resend(sender, request, body_producer):
+            sender.agent = _get_agent()
+            if sender.gssclient is not None:
+                # do some cleanup first.  memory leaks were occurring
+                sender.gssclient.cleanup()
+                sender.gssclient = None
+                try:
+                    yield sender._set_url_and_headers()
+                    encrypted_request = sender.gssclient.encrypt_body(request)
+                    if not encrypted_request.startswith("--Encrypted Boundary"):
+                        sender._headers.setRawHeaders('Content-Type', _CONTENT_TYPE['Content-Type'])
+                    body_producer = _StringProducer(encrypted_request)
+                except Exception as e:
+                    raise e
+            try:
+                response = yield sender.agent.request(
+                    'POST', sender._url, sender._headers, body_producer)
+            except Exception as e:
+                raise e
+            defer.returnValue(response)
+
         try:
             response = yield self.agent.request(
                 'POST', self._url, self._headers, body_producer)
+        except ConnectError:
+            # network timeout.  could be stale connection, so let's reset
+            response = yield reset_agent_resend(self, request, body_producer)
         except Exception as e:
             raise e
+
         log.debug('received response {0} {1}'.format(
             response.code, request_template_name))
         if response.code == httplib.UNAUTHORIZED or response.code == httplib.BAD_REQUEST:
             # check to see if we need to re-authorize due to lost connection or bad request error
-            if self.gssclient is not None:
-                self.agent = _get_agent()
-                # do some cleanup first.  memory leaks were occurring
-                self.gssclient.cleanup()
-                self.gssclient = None
-                try:
-                    yield self._set_url_and_headers()
-                    encrypted_request = self.gssclient.encrypt_body(request)
-                    if not encrypted_request.startswith("--Encrypted Boundary"):
-                        self._headers.setRawHeaders('Content-Type', _CONTENT_TYPE['Content-Type'])
-                    body_producer = _StringProducer(encrypted_request)
-                    response = yield self.agent.request(
-                        'POST', self._url, self._headers, body_producer)
-                except Exception as e:
-                    raise e
+            response = yield reset_agent_resend(self, request, body_producer)
             if response.code == httplib.UNAUTHORIZED:
                 if self.is_kerberos():
                     auth_header = response.headers.getRawHeaders('WWW-Authenticate')[0]
